@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2015-2018 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2015-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -21,8 +21,10 @@
 #ifndef BFS_BFTW_H
 #define BFS_BFTW_H
 
+#include "mtab.h"
 #include "stat.h"
 #include <stddef.h>
+#include <sys/types.h>
 
 /**
  * Possible file types.
@@ -69,6 +71,18 @@ enum bftw_visit {
 	BFTW_POST,
 };
 
+/**
+ * Cached bfs_stat() info for a file.
+ */
+struct bftw_stat {
+	/** A pointer to the bfs_stat() buffer, if available. */
+	const struct bfs_stat *buf;
+	/** Storage for the bfs_stat() buffer, if needed. */
+	struct bfs_stat storage;
+	/** The cached error code, if any. */
+	int error;
+};
+
 /**
  * Data about the current file for the bftw() callback.
  */
@@ -90,27 +104,54 @@ struct BFTW {
 	/** The errno that occurred, if typeflag == BFTW_ERROR. */
 	int error;
 
-	/** A bfs_stat() buffer; may be NULL if no stat() call was needed. */
-	const struct bfs_stat *statbuf;
-
 	/** A parent file descriptor for the *at() family of calls. */
 	int at_fd;
-	/** The path relative to atfd for the *at() family of calls. */
+	/** The path relative to at_fd for the *at() family of calls. */
 	const char *at_path;
-	/** Appropriate flags (such as AT_SYMLINK_NOFOLLOW) for the *at() family of calls. */
-	int at_flags;
+
+	/** Flags for bfs_stat(). */
+	enum bfs_stat_flag stat_flags;
+	/** Cached bfs_stat() info for BFS_STAT_NOFOLLOW. */
+	struct bftw_stat lstat_cache;
+	/** Cached bfs_stat() info for BFS_STAT_FOLLOW. */
+	struct bftw_stat stat_cache;
 };
 
+/**
+ * Get bfs_stat() info for a file encountered during bftw(), caching the result
+ * whenever possible.
+ *
+ * @param ftwbuf
+ *         bftw() data for the file to stat.
+ * @param flags
+ *         flags for bfs_stat().  Pass ftwbuf->stat_flags for the default flags.
+ * @return
+ *         A pointer to a bfs_stat() buffer, or NULL if the call failed.
+ */
+const struct bfs_stat *bftw_stat(const struct BFTW *ftwbuf, enum bfs_stat_flag flags);
+
+/**
+ * Get the type of a file encountered during bftw(), with flags controlling
+ * whether to follow links.  This function will avoid calling bfs_stat() if
+ * possible.
+ *
+ * @param ftwbuf
+ *         bftw() data for the file to check.
+ * @param flags
+ *         flags for bfs_stat().  Pass ftwbuf->stat_flags for the default flags.
+ * @return
+ *         The type of the file, or BFTW_ERROR if an error occurred.
+ */
+enum bftw_typeflag bftw_typeflag(const struct BFTW *ftwbuf, enum bfs_stat_flag flags);
+
 /**
  * Walk actions returned by the bftw() callback.
  */
 enum bftw_action {
 	/** Keep walking. */
 	BFTW_CONTINUE,
-	/** Skip this path's siblings. */
-	BFTW_SKIP_SIBLINGS,
 	/** Skip this path's children. */
-	BFTW_SKIP_SUBTREE,
+	BFTW_PRUNE,
 	/** Stop walking. */
 	BFTW_STOP,
 };
@@ -125,7 +166,7 @@ enum bftw_action {
  * @return
  *         An action value.
  */
-typedef enum bftw_action bftw_fn(struct BFTW *ftwbuf, void *ptr);
+typedef enum bftw_action bftw_callback(const struct BFTW *ftwbuf, void *ptr);
 
 /**
  * Flags that control bftw() behavior.
@@ -147,26 +188,51 @@ enum bftw_flags {
 	BFTW_XDEV          = 1 << 6,
 };
 
+/**
+ * Tree search strategies for bftw().
+ */
+enum bftw_strategy {
+	/** Breadth-first search. */
+	BFTW_BFS,
+	/** Depth-first search. */
+	BFTW_DFS,
+	/** Iterative deepening search. */
+	BFTW_IDS,
+};
+
+/**
+ * Structure for holding the arguments passed to bftw().
+ */
+struct bftw_args {
+	/** The path(s) to start from. */
+	const char **paths;
+	/** The number of starting paths. */
+	size_t npaths;
+	/** The callback to invoke. */
+	bftw_callback *callback;
+	/** A pointer which is passed to the callback. */
+	void *ptr;
+	/** The maximum number of file descriptors to keep open. */
+	int nopenfd;
+	/** Flags that control bftw() behaviour. */
+	enum bftw_flags flags;
+	/** The search strategy to use. */
+	enum bftw_strategy strategy;
+	/** The parsed mount table, if available. */
+	const struct bfs_mtab *mtab;
+};
+
 /**
  * Breadth First Tree Walk (or Better File Tree Walk).
  *
  * Like ftw(3) and nftw(3), this function walks a directory tree recursively,
- * and invokes a callback for each path it encounters.  However, bftw() operates
- * breadth-first.
+ * and invokes a callback for each path it encounters.
  *
- * @param path
- *         The starting path.
- * @param fn
- *         The callback to invoke.
- * @param nopenfd
- *         The maximum number of file descriptors to keep open.
- * @param flags
- *         Flags that control bftw() behavior.
- * @param ptr
- *         A generic pointer which is passed to fn().
+ * @param args
+ *         The arguments that control the walk.
  * @return
  *         0 on success, or -1 on failure.
  */
-int bftw(const char *path, bftw_fn *fn, int nopenfd, enum bftw_flags flags, void *ptr);
+int bftw(const struct bftw_args *args);
 
 #endif // BFS_BFTW_H
