@@ -22,6 +22,7 @@
 #define BFS_CMDLINE_H
 
 #include "color.h"
+#include "trie.h"
 
 /**
  * Various debugging flags.
@@ -45,21 +46,6 @@ enum debug_flags {
 	DEBUG_ALL    = (1 << 7) - 1,
 };
 
-/**
- * A root path to explore.
- */
-struct root {
-	/** The root path itself. */
-	const char *path;
-	/** The next path in the list. */
-	struct root *next;
-};
-
-/**
- * An open file for the command line.
- */
-struct open_file;
-
 /**
  * The parsed command line.
  */
@@ -67,8 +53,10 @@ struct cmdline {
 	/** The unparsed command line arguments. */
 	char **argv;
 
-	/** The list of root paths. */
-	struct root *roots;
+	/** The root paths. */
+	const char **paths;
+	/** The number of root paths. */
+	size_t npaths;
 
 	/** Color data. */
 	struct colors *colors;
@@ -79,6 +67,8 @@ struct cmdline {
 
 	/** Table of mounted file systems. */
 	struct bfs_mtab *mtab;
+	/** The error that occurred parsing the mount table, if any. */
+	int mtab_error;
 
 	/** -mindepth option. */
 	int mindepth;
@@ -87,6 +77,8 @@ struct cmdline {
 
 	/** bftw() flags. */
 	enum bftw_flags flags;
+	/** bftw() search strategy. */
+	enum bftw_strategy strategy;
 
 	/** Optimization level. */
 	int optlevel;
@@ -96,12 +88,14 @@ struct cmdline {
 	bool xargs_safe;
 	/** Whether to ignore deletions that race with bfs. */
 	bool ignore_races;
+	/** Whether to only return unique files. */
+	bool unique;
 
 	/** The command line expression. */
 	struct expr *expr;
 
 	/** All the open files owned by the command line. */
-	struct open_file *open_files;
+	struct trie open_files;
 	/** The number of open files owned by the command line. */
 	int nopen_files;
 };
