@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2017 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -15,6 +15,9 @@
  ****************************************************************************/
 
 #include "dstring.h"
+#include <assert.h>
+#include <stdarg.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -37,16 +40,25 @@ static size_t dstrsize(size_t capacity) {
 	return sizeof(struct dstring) + capacity + 1;
 }
 
-char *dstralloc(size_t capacity) {
+/** Allocate a dstring with the given contents. */
+static char *dstralloc_impl(size_t capacity, size_t length, const char *data) {
 	struct dstring *header = malloc(dstrsize(capacity));
 	if (!header) {
 		return NULL;
 	}
 
 	header->capacity = capacity;
-	header->length = 0;
-	header->data[0] = '\0';
-	return header->data;
+	header->length = length;
+	return memcpy(header->data, data, length + 1);
+}
+
+char *dstralloc(size_t capacity) {
+	return dstralloc_impl(capacity, 0, "");
+}
+
+char *dstrdup(const char *str) {
+	size_t len = strlen(str);
+	return dstralloc_impl(len, len, str);
 }
 
 size_t dstrlen(const char *dstr) {
@@ -108,6 +120,31 @@ int dstrapp(char **str, char c) {
 	return dstrcat_impl(str, &c, 1);
 }
 
+char *dstrprintf(const char *format, ...) {
+	va_list args;
+
+	va_start(args, format);
+	int len = vsnprintf(NULL, 0, format, args);
+	va_end(args);
+
+	assert(len > 0);
+
+	char *str = dstralloc(len);
+	if (!str) {
+		return NULL;
+	}
+
+	va_start(args, format);
+	len = vsnprintf(str, len + 1, format, args);
+	va_end(args);
+
+	struct dstring *header = dstrheader(str);
+	assert(len == header->capacity);
+	header->length = len;
+
+	return str;
+}
+
 void dstrfree(char *dstr) {
 	if (dstr) {
 		free(dstrheader(dstr));
