@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2016-2017 Tavian Barnes <tavianator@tavianator.com>        *
+ * Copyright (C) 2016-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -21,6 +21,7 @@
 #ifndef BFS_DSTRING_H
 #define BFS_DSTRING_H
 
+#include "util.h"
 #include <stddef.h>
 
 /**
@@ -31,6 +32,14 @@
  */
 char *dstralloc(size_t capacity);
 
+/**
+ * Create a dynamic copy of a string.
+ *
+ * @param str
+ *         The NUL-terminated string to copy.
+ */
+char *dstrdup(const char *str);
+
 /**
  * Get a dynamic string's length.
  *
@@ -97,6 +106,19 @@ int dstrncat(char **dest, const char *src, size_t n);
  */
 int dstrapp(char **str, char c);
 
+/**
+ * Create a dynamic string from a format string.
+ *
+ * @param format
+ *         The format string to fill in.
+ * @param ...
+ *         Any arguments for the format string.
+ * @return
+ *         The created string, or NULL on failure.
+ */
+BFS_FORMATTER(1, 2)
+char *dstrprintf(const char *format, ...);
+
 /**
  * Free a dynamic string.
  *
