@@ -28,12 +28,14 @@
 #include "eval.h"
 #include "exec.h"
 #include "expr.h"
+#include "fsade.h"
 #include "mtab.h"
-#include "posix1e.h"
 #include "printf.h"
+#include "spawn.h"
 #include "stat.h"
 #include "typo.h"
 #include "util.h"
+#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -49,6 +51,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -246,12 +249,6 @@ struct open_file {
 	CFILE *cfile;
 	/** The path to the file (for diagnostics). */
 	const char *path;
-	/** Device number (for deduplication). */
-	dev_t dev;
-	/** Inode number (for deduplication). */
-	ino_t ino;
-	/** The next open file in the chain. */
-	struct open_file *next;
 };
 
 /**
@@ -268,9 +265,9 @@ int free_cmdline(struct cmdline *cmdline) {
 
 		free_bfs_mtab(cmdline->mtab);
 
-		struct open_file *ofile = cmdline->open_files;
-		while (ofile) {
-			struct open_file *next = ofile->next;
+		struct trie_leaf *leaf;
+		while ((leaf = trie_first_leaf(&cmdline->open_files))) {
+			struct open_file *ofile = leaf->value;
 
 			if (cfclose(ofile->cfile) != 0) {
 				if (cerr) {
@@ -280,8 +277,9 @@ int free_cmdline(struct cmdline *cmdline) {
 			}
 
 			free(ofile);
-			ofile = next;
+			trie_remove(&cmdline->open_files, leaf);
 		}
+		trie_destroy(&cmdline->open_files);
 
 		if (cout && fflush(cout->file) != 0) {
 			if (cerr) {
@@ -294,14 +292,7 @@ int free_cmdline(struct cmdline *cmdline) {
 		cfclose(cerr);
 
 		free_colors(cmdline->colors);
-
-		struct root *root = cmdline->roots;
-		while (root) {
-			struct root *next = root->next;
-			free(root);
-			root = next;
-		}
-
+		free(cmdline->paths);
 		free(cmdline->argv);
 		free(cmdline);
 	}
@@ -328,13 +319,13 @@ struct parser_state {
 	char **argv;
 	/** The name of this program. */
 	const char *command;
-	/** The current tail of the root path list. */
-	struct root **roots_tail;
 
 	/** The current regex flags to use. */
 	int regex_flags;
 
-	/** Whether this session is interactive. */
+	/** Whether stdout is a terminal. */
+	bool stdout_tty;
+	/** Whether this session is interactive (stdin and stderr are each a terminal). */
 	bool interactive;
 	/** Whether -color or -nocolor has been passed. */
 	enum use_color use_color;
@@ -424,31 +415,32 @@ static int expr_open(struct parser_state *state, struct expr *expr, const char *
 	}
 
 	struct bfs_stat sb;
-	if (bfs_fstat(fileno(cfile->file), &sb) != 0) {
+	if (bfs_stat(fileno(cfile->file), NULL, 0, &sb) != 0) {
 		parse_error(state, "'%s': %m.\n", path);
 		goto out_close;
 	}
 
-	for (struct open_file *ofile = cmdline->open_files; ofile; ofile = ofile->next) {
-		if (ofile->dev == sb.dev && ofile->ino == sb.ino) {
-			expr->cfile = ofile->cfile;
-			ret = 0;
-			goto out_close;
-		}
+	bfs_file_id id;
+	bfs_stat_id(&sb, &id);
+
+	struct trie_leaf *leaf = trie_insert_mem(&cmdline->open_files, id, sizeof(id));
+	if (leaf->value) {
+		struct open_file *ofile = leaf->value;
+		expr->cfile = ofile->cfile;
+		ret = 0;
+		goto out_close;
 	}
 
 	struct open_file *ofile = malloc(sizeof(*ofile));
 	if (!ofile) {
 		perror("malloc()");
+		trie_remove(&cmdline->open_files, leaf);
 		goto out_close;
 	}
 
 	ofile->cfile = cfile;
 	ofile->path = path;
-	ofile->dev = sb.dev;
-	ofile->ino = sb.ino;
-	ofile->next = cmdline->open_files;
-	cmdline->open_files = ofile;
+	leaf->value = ofile;
 	++cmdline->nopen_files;
 
 	expr->cfile = cfile;
@@ -469,9 +461,9 @@ static int stat_arg(const struct parser_state *state, struct expr *expr, struct
 	const struct cmdline *cmdline = state->cmdline;
 
 	bool follow = cmdline->flags & (BFTW_COMFOLLOW | BFTW_LOGICAL);
-	int at_flags = follow ? 0 : AT_SYMLINK_NOFOLLOW;
+	enum bfs_stat_flag flags = follow ? BFS_STAT_TRYFOLLOW : BFS_STAT_NOFOLLOW;
 
-	int ret = bfs_stat(AT_FDCWD, expr->sdata, at_flags, BFS_STAT_BROKEN_OK, sb);
+	int ret = bfs_stat(AT_FDCWD, expr->sdata, flags, sb);
 	if (ret != 0) {
 		parse_error(state, "'%s': %m.\n", expr->sdata);
 	}
@@ -508,16 +500,18 @@ static char **parser_advance(struct parser_state *state, enum token_type type, s
  * Parse a root path.
  */
 static int parse_root(struct parser_state *state, const char *path) {
-	struct root *root = malloc(sizeof(struct root));
-	if (!root) {
-		perror("malloc()");
-		return -1;
+	struct cmdline *cmdline = state->cmdline;
+	size_t i = cmdline->npaths;
+	if ((i & (i + 1)) == 0) {
+		const char **paths = realloc(cmdline->paths, 2*(i + 1)*sizeof(*paths));
+		if (!paths) {
+			return -1;
+		}
+		cmdline->paths = paths;
 	}
 
-	root->path = path;
-	root->next = NULL;
-	*state->roots_tail = root;
-	state->roots_tail = &root->next;
+	cmdline->paths[i] = path;
+	cmdline->npaths = i + 1;
 	return 0;
 }
 
@@ -662,10 +656,16 @@ static const char *parse_icmp(const struct parser_state *state, const char *str,
  * Check if a string could be an integer comparison.
  */
 static bool looks_like_icmp(const char *str) {
-	if (str[0] == '-' || str[0] == '+') {
-		++str;
+	int i;
+
+	// One +/- for the comparison flag, one for the sign
+	for (i = 0; i < 2; ++i) {
+		if (str[i] != '-' && str[i] != '+') {
+			break;
+		}
 	}
-	return str[0] >= '0' && str[0] <= '9';
+
+	return str[i] >= '0' && str[i] <= '9';
 }
 
 /**
@@ -965,7 +965,7 @@ static struct expr *parse_optlevel(struct parser_state *state, int arg1, int arg
  */
 static struct expr *parse_follow(struct parser_state *state, int flags, int option) {
 	struct cmdline *cmdline = state->cmdline;
-	cmdline->flags &= ~(BFTW_COMFOLLOW | BFTW_LOGICAL | BFTW_DETECT_CYCLES);
+	cmdline->flags &= ~(BFTW_COMFOLLOW | BFTW_LOGICAL);
 	cmdline->flags |= flags;
 	if (option) {
 		return parse_nullary_positional_option(state);
@@ -987,9 +987,25 @@ static struct expr *parse_xargs_safe(struct parser_state *state, int arg1, int a
  */
 static struct expr *parse_access(struct parser_state *state, int flag, int arg2) {
 	struct expr *expr = parse_nullary_test(state, eval_access);
-	if (expr) {
-		expr->idata = flag;
+	if (!expr) {
+		return NULL;
 	}
+
+	expr->idata = flag;
+	expr->cost = STAT_COST;
+
+	switch (flag) {
+	case R_OK:
+		expr->probability = 0.99;
+		break;
+	case W_OK:
+		expr->probability = 0.8;
+		break;
+	case X_OK:
+		expr->probability = 0.2;
+		break;
+	}
+
 	return expr;
 }
 
@@ -997,12 +1013,11 @@ static struct expr *parse_access(struct parser_state *state, int flag, int arg2)
  * Parse -acl.
  */
 static struct expr *parse_acl(struct parser_state *state, int flag, int arg2) {
-#if BFS_HAS_SYS_ACL
+#if BFS_CAN_CHECK_ACL
 	struct expr *expr = parse_nullary_test(state, eval_acl);
 	if (expr) {
 		expr->cost = STAT_COST;
 		expr->probability = 0.00002;
-		expr->ephemeral_fds = 1;
 	}
 	return expr;
 #else
@@ -1055,12 +1070,11 @@ static struct expr *parse_time(struct parser_state *state, int field, int unit)
  * Parse -capable.
  */
 static struct expr *parse_capable(struct parser_state *state, int flag, int arg2) {
-#if BFS_HAS_POSIX1E_CAPABILITIES
+#if BFS_CAN_CHECK_CAPABILITIES
 	struct expr *expr = parse_nullary_test(state, eval_capable);
 	if (expr) {
 		expr->cost = STAT_COST;
 		expr->probability = 0.000002;
-		expr->ephemeral_fds = 1;
 	}
 	return expr;
 #else
@@ -1380,11 +1394,8 @@ fail:
 static struct expr *parse_fstype(struct parser_state *state, int arg1, int arg2) {
 	struct cmdline *cmdline = state->cmdline;
 	if (!cmdline->mtab) {
-		cmdline->mtab = parse_bfs_mtab();
-		if (!cmdline->mtab) {
-			parse_error(state, "Couldn't parse the mount table: %m.\n");
-			return NULL;
-		}
+		parse_error(state, "Couldn't parse the mount table: %s.\n", strerror(cmdline->mtab_error));
+		return NULL;
 	}
 
 	struct expr *expr = parse_unary_test(state, eval_fstype);
@@ -1427,6 +1438,14 @@ fail:
 	return NULL;
 }
 
+/**
+ * Parse -unique.
+ */
+static struct expr *parse_unique(struct parser_state *state, int arg1, int arg2) {
+	state->cmdline->unique = true;
+	return parse_nullary_option(state);
+}
+
 /**
  * Parse -used N.
  */
@@ -1651,6 +1670,8 @@ static struct expr *parse_newerxy(struct parser_state *state, int arg1, int arg2
 		expr->reftime = *reftime;
 	}
 
+	expr->cost = STAT_COST;
+
 	return expr;
 
 fail:
@@ -1665,6 +1686,7 @@ static struct expr *parse_nogroup(struct parser_state *state, int arg1, int arg2
 	struct expr *expr = parse_nullary_test(state, eval_nogroup);
 	if (expr) {
 		expr->cost = 9000.0;
+		expr->probability = 0.01;
 		expr->ephemeral_fds = 1;
 	}
 	return expr;
@@ -1696,6 +1718,7 @@ static struct expr *parse_nouser(struct parser_state *state, int arg1, int arg2)
 	struct expr *expr = parse_nullary_test(state, eval_nouser);
 	if (expr) {
 		expr->cost = 9000.0;
+		expr->probability = 0.01;
 		expr->ephemeral_fds = 1;
 	}
 	return expr;
@@ -2147,6 +2170,46 @@ static struct expr *parse_samefile(struct parser_state *state, int arg1, int arg
 	return expr;
 }
 
+/**
+ * Parse -S STRATEGY.
+ */
+static struct expr *parse_search_strategy(struct parser_state *state, int arg1, int arg2) {
+	const char *flag = state->argv[0];
+	const char *arg = state->argv[1];
+	if (!arg) {
+		parse_error(state, "%s needs an argument.\n", flag);
+		return NULL;
+	}
+
+	struct cmdline *cmdline = state->cmdline;
+	FILE *file = stderr;
+
+	if (strcmp(arg, "bfs") == 0) {
+		cmdline->strategy = BFTW_BFS;
+	} else if (strcmp(arg, "dfs") == 0) {
+		cmdline->strategy = BFTW_DFS;
+	} else if (strcmp(arg, "ids") == 0) {
+		cmdline->strategy = BFTW_IDS;
+	} else if (strcmp(arg, "help") == 0) {
+		state->just_info = true;
+		file = stdout;
+		goto fail_list_strategies;
+	} else {
+		goto fail_bad_strategy;
+	}
+
+	return parse_unary_flag(state);
+
+fail_bad_strategy:
+	parse_error(state, "Unrecognized search strategy '%s'.\n\n", arg);
+fail_list_strategies:
+	fputs("Supported search strategies:\n\n", file);
+	fputs("  bfs: breadth-first search\n", file);
+	fputs("  dfs: depth-first search\n", file);
+	fputs("  ids: iterative deepening search\n", file);
+	return NULL;
+}
+
 /**
  * Parse -size N[cwbkMGTP]?.
  */
@@ -2328,11 +2391,139 @@ static struct expr *parse_warn(struct parser_state *state, int warn, int arg2) {
 	return parse_nullary_positional_option(state);
 }
 
+/**
+ * Parse -xattr.
+ */
+static struct expr *parse_xattr(struct parser_state *state, int arg1, int arg2) {
+#if BFS_CAN_CHECK_XATTRS
+	struct expr *expr = parse_nullary_test(state, eval_xattr);
+	if (expr) {
+		expr->cost = STAT_COST;
+		expr->probability = 0.01;
+	}
+	return expr;
+#else
+	parse_error(state, "%s is missing platform support.\n", state->argv[0]);
+	return NULL;
+#endif
+}
+
+/**
+ * Launch a pager for the help output.
+ */
+static CFILE *launch_pager(pid_t *pid, CFILE *cout) {
+	char *pager = getenv("PAGER");
+	if (!pager) {
+		pager = "more";
+	}
+
+	int pipefd[2];
+	if (pipe(pipefd) != 0) {
+		goto fail;
+	}
+
+	FILE *file = fdopen(pipefd[1], "w");
+	if (!file) {
+		goto fail_pipe;
+	}
+	pipefd[1] = -1;
+
+	CFILE *ret = cfdup(file, NULL);
+	if (!ret) {
+		goto fail_file;
+	}
+	file = NULL;
+	ret->close = true;
+	ret->colors = cout->colors;
+
+	struct bfs_spawn ctx;
+	if (bfs_spawn_init(&ctx) != 0) {
+		goto fail_ret;
+	}
+
+	if (bfs_spawn_setflags(&ctx, BFS_SPAWN_USEPATH) != 0) {
+		goto fail_ctx;
+	}
+
+	if (bfs_spawn_addclose(&ctx, fileno(ret->file)) != 0) {
+		goto fail_ctx;
+	}
+	if (bfs_spawn_adddup2(&ctx, pipefd[0], STDIN_FILENO) != 0) {
+		goto fail_ctx;
+	}
+	if (bfs_spawn_addclose(&ctx, pipefd[0]) != 0) {
+		goto fail_ctx;
+	}
+
+	char *argv[] = {
+		pager,
+		NULL,
+	};
+
+	extern char **environ;
+	char **envp = environ;
+
+	if (!getenv("LESS")) {
+		size_t envc;
+		for (envc = 0; environ[envc]; ++envc);
+		++envc;
+
+		envp = malloc((envc + 1)*sizeof(*envp));
+		if (!envp) {
+			goto fail_ctx;
+		}
+
+		memcpy(envp, environ, (envc - 1)*sizeof(*envp));
+		envp[envc - 1] = "LESS=FKRX";
+		envp[envc] = NULL;
+	}
+
+	*pid = bfs_spawn(pager, &ctx, argv, envp);
+	if (*pid < 0) {
+		goto fail_envp;
+	}
+
+	close(pipefd[0]);
+	if (envp != environ) {
+		free(envp);
+	}
+	bfs_spawn_destroy(&ctx);
+	return ret;
+
+fail_envp:
+	if (envp != environ) {
+		free(envp);
+	}
+fail_ctx:
+	bfs_spawn_destroy(&ctx);
+fail_ret:
+	cfclose(ret);
+fail_file:
+	if (file) {
+		fclose(file);
+	}
+fail_pipe:
+	if (pipefd[1] >= 0) {
+		close(pipefd[1]);
+	}
+	if (pipefd[0] >= 0) {
+		close(pipefd[0]);
+	}
+fail:
+	return cout;
+}
+
 /**
  * "Parse" -help.
  */
 static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	CFILE *cout = state->cmdline->cout;
+
+	pid_t pager = -1;
+	if (state->stdout_tty) {
+		cout = launch_pager(&pager, cout);
+	}
+
 	cfprintf(cout, "Usage: ${ex}%s${rs} [${cyn}flags${rs}...] [${mag}paths${rs}...] [${blu}expression${rs}...]\n\n",
 		 state->command);
 
@@ -2363,7 +2554,9 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${cyn}-D${rs} ${bld}FLAG${rs}\n");
 	cfprintf(cout, "      Turn on a debugging flag (see ${cyn}-D${rs} ${bld}help${rs})\n");
 	cfprintf(cout, "  ${cyn}-O${rs}${bld}N${rs}\n");
-	cfprintf(cout, "      Enable optimization level ${bld}N${rs} (default: 3)\n\n");
+	cfprintf(cout, "      Enable optimization level ${bld}N${rs} (default: 3)\n");
+	cfprintf(cout, "  ${cyn}-S${rs} ${bld}bfs${rs}|${bld}dfs${rs}|${bld}ids${rs}\n");
+	cfprintf(cout, "      Use ${bld}b${rs}readth-${bld}f${rs}irst/${bld}d${rs}epth-${bld}f${rs}irst/${bld}i${rs}terative ${bld}d${rs}eepening ${bld}s${rs}earch (default: ${cyn}-S${rs} ${bld}bfs${rs})\n\n");
 
 	cfprintf(cout, "${bld}Operators:${rs}\n\n");
 
@@ -2408,25 +2601,28 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${blu}-regextype${rs} ${bld}TYPE${rs}\n");
 	cfprintf(cout, "      Use ${bld}TYPE${rs}-flavored regexes (default: ${bld}posix-basic${rs}; see ${blu}-regextype${rs}"
 	                " ${bld}help${rs})\n");
+	cfprintf(cout, "  ${blu}-unique${rs}\n");
+	cfprintf(cout, "      Skip any files that have already been seen\n");
 	cfprintf(cout, "  ${blu}-warn${rs}\n");
 	cfprintf(cout, "  ${blu}-nowarn${rs}\n");
 	cfprintf(cout, "      Turn on or off warnings about the command line\n\n");
 
 	cfprintf(cout, "${bld}Tests:${rs}\n\n");
 
-#if BFS_HAS_SYS_ACL
+#if BFS_CAN_CHECK_ACL
 	cfprintf(cout, "  ${blu}-acl${rs}\n");
-	cfprintf(cout, "      Find files with non-trivial Access Control Lists\n");
+	cfprintf(cout, "      Find files with a non-trivial Access Control List\n");
 #endif
 	cfprintf(cout, "  ${blu}-${rs}[${blu}aBcm${rs}]${blu}min${rs} ${bld}[-+]N${rs}\n");
 	cfprintf(cout, "      Find files ${blu}a${rs}ccessed/${blu}B${rs}irthed/${blu}c${rs}hanged/${blu}m${rs}odified ${bld}N${rs} minutes ago\n");
 	cfprintf(cout, "  ${blu}-${rs}[${blu}aBcm${rs}]${blu}newer${rs} ${bld}FILE${rs}\n");
-	cfprintf(cout, "      Find files ${blu}a${rs}ccessed/${blu}B${rs}irthed/${blu}c${rs}hanged/${blu}m${rs}odified more recently than ${bld}FILE${rs} was modified\n");
+	cfprintf(cout, "      Find files ${blu}a${rs}ccessed/${blu}B${rs}irthed/${blu}c${rs}hanged/${blu}m${rs}odified more recently than ${bld}FILE${rs} was\n"
+	               "      modified\n");
 	cfprintf(cout, "  ${blu}-${rs}[${blu}aBcm${rs}]${blu}time${rs} ${bld}[-+]N${rs}\n");
 	cfprintf(cout, "      Find files ${blu}a${rs}ccessed/${blu}B${rs}irthed/${blu}c${rs}hanged/${blu}m${rs}odified ${bld}N${rs} days ago\n");
-#if BFS_HAS_POSIX1E_CAPABILITIES
+#if BFS_CAN_CHECK_CAPABILITIES
 	cfprintf(cout, "  ${blu}-capable${rs}\n");
-	cfprintf(cout, "      Match files with POSIX.1e capabilities set\n");
+	cfprintf(cout, "      Find files with POSIX.1e capabilities set\n");
 #endif
 	cfprintf(cout, "  ${blu}-depth${rs} ${bld}[-+]N${rs}\n");
 	cfprintf(cout, "      Find files with depth ${bld}N${rs}\n");
@@ -2449,7 +2645,7 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "      Find files owned by the group/user ${bld}NAME${rs}\n");
 	cfprintf(cout, "  ${blu}-hidden${rs}\n");
 	cfprintf(cout, "  ${blu}-nohidden${rs}\n");
-	cfprintf(cout, "      Match hidden files, or filter them out\n");
+	cfprintf(cout, "      Find hidden files, or filter them out\n");
 #ifdef FNM_CASEFOLD
 	cfprintf(cout, "  ${blu}-ilname${rs} ${bld}GLOB${rs}\n");
 	cfprintf(cout, "  ${blu}-iname${rs}  ${bld}GLOB${rs}\n");
@@ -2477,7 +2673,7 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${blu}-nouser${rs}\n");
 	cfprintf(cout, "      Find files owned by nonexistent groups/users\n");
 	cfprintf(cout, "  ${blu}-path${rs} ${bld}GLOB${rs}\n");
-	cfprintf(cout, "  ${blu}-wholename{rs} ${bld}GLOB${rs}\n");
+	cfprintf(cout, "  ${blu}-wholename${rs} ${bld}GLOB${rs}\n");
 	cfprintf(cout, "      Find files whose entire path matches the ${bld}GLOB${rs}\n");
 	cfprintf(cout, "  ${blu}-perm${rs} ${bld}[-]MODE${rs}\n");
 	cfprintf(cout, "      Find files with a matching mode\n");
@@ -2494,6 +2690,10 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "      Find files of the given type\n");
 	cfprintf(cout, "  ${blu}-used${rs} ${bld}[-+]N${rs}\n");
 	cfprintf(cout, "      Find files last accessed ${bld}N${rs} days after they were changed\n");
+#if BFS_CAN_CHECK_XATTRS
+	cfprintf(cout, "  ${blu}-xattr${rs}\n");
+	cfprintf(cout, "      Find files with extended attributes\n");
+#endif
 	cfprintf(cout, "  ${blu}-xtype${rs} ${bld}[bcdlpfswD]${rs}\n");
 	cfprintf(cout, "      Find files of the given type, following links when ${blu}-type${rs} would not, and\n");
 	cfprintf(cout, "      vice versa\n\n");
@@ -2520,8 +2720,8 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 	cfprintf(cout, "  ${blu}-fprint${rs} ${bld}FILE${rs}\n");
 	cfprintf(cout, "  ${blu}-fprint0${rs} ${bld}FILE${rs}\n");
 	cfprintf(cout, "  ${blu}-fprintf${rs} ${bld}FORMAT${rs} ${bld}FILE${rs}\n");
-	cfprintf(cout, "      Like ${blu}-ls${rs}/${blu}-print${rs}/${blu}-print0${rs}/${blu}-printf${rs}, but write to"
-	               " ${bld}FILE${rs} instead of standard output\n");
+	cfprintf(cout, "      Like ${blu}-ls${rs}/${blu}-print${rs}/${blu}-print0${rs}/${blu}-printf${rs}, but write to ${bld}FILE${rs} instead of standard\n"
+	               "      output\n");
 	cfprintf(cout, "  ${blu}-ls${rs}\n");
 	cfprintf(cout, "      List files like ${ex}ls${rs} ${bld}-dils${rs}\n");
 	cfprintf(cout, "  ${blu}-nohidden${rs}\n");
@@ -2548,6 +2748,11 @@ static struct expr *parse_help(struct parser_state *state, int arg1, int arg2) {
 
 	cfprintf(cout, "%s\n", BFS_HOMEPAGE);
 
+	if (pager > 0) {
+		cfclose(cout);
+		waitpid(pager, NULL, 0);
+	}
+
 	state->just_info = true;
 	return NULL;
 }
@@ -2571,119 +2776,122 @@ typedef struct expr *parse_fn(struct parser_state *state, int arg1, int arg2);
  */
 struct table_entry {
 	char *arg;
-	bool prefix;
 	parse_fn *parse;
 	int arg1;
 	int arg2;
+	bool prefix;
 };
 
 /**
  * The parse table for literals.
  */
 static const struct table_entry parse_table[] = {
-	{"-Bmin", false, parse_time, BFS_STAT_BTIME, MINUTES},
-	{"-Bnewer", false, parse_newer, BFS_STAT_BTIME},
-	{"-Btime", false, parse_time, BFS_STAT_BTIME, DAYS},
-	{"-D", false, parse_debug},
-	{"-E", false, parse_regex_extended},
-	{"-O", true, parse_optlevel},
-	{"-P", false, parse_follow, 0, false},
-	{"-H", false, parse_follow, BFTW_COMFOLLOW, false},
-	{"-L", false, parse_follow, BFTW_LOGICAL | BFTW_DETECT_CYCLES, false},
-	{"-X", false, parse_xargs_safe},
-	{"-a"},
-	{"-acl", false, parse_acl},
-	{"-amin", false, parse_time, BFS_STAT_ATIME, MINUTES},
-	{"-and"},
-	{"-anewer", false, parse_newer, BFS_STAT_ATIME},
-	{"-atime", false, parse_time, BFS_STAT_ATIME, DAYS},
-	{"-capable", false, parse_capable},
-	{"-cmin", false, parse_time, BFS_STAT_CTIME, MINUTES},
-	{"-cnewer", false, parse_newer, BFS_STAT_CTIME},
-	{"-ctime", false, parse_time, BFS_STAT_CTIME, DAYS},
-	{"-color", false, parse_color, true},
-	{"-d", false, parse_depth},
-	{"-daystart", false, parse_daystart},
-	{"-delete", false, parse_delete},
-	{"-depth", false, parse_depth_n},
-	{"-empty", false, parse_empty},
-	{"-exec", false, parse_exec, 0},
-	{"-execdir", false, parse_exec, BFS_EXEC_CHDIR},
-	{"-executable", false, parse_access, X_OK},
-	{"-exit", false, parse_exit},
-	{"-f", false, parse_f},
-	{"-false", false, parse_const, false},
-	{"-fls", false, parse_fls},
-	{"-follow", false, parse_follow, BFTW_LOGICAL | BFTW_DETECT_CYCLES, true},
-	{"-fprint", false, parse_fprint},
-	{"-fprint0", false, parse_fprint0},
-	{"-fprintf", false, parse_fprintf},
-	{"-fstype", false, parse_fstype},
-	{"-gid", false, parse_group},
-	{"-group", false, parse_group},
-	{"-help", false, parse_help},
-	{"-hidden", false, parse_hidden},
-	{"-ignore_readdir_race", false, parse_ignore_races, true},
-	{"-ilname", false, parse_lname, true},
-	{"-iname", false, parse_name, true},
-	{"-inum", false, parse_inum},
-	{"-ipath", false, parse_path, true},
-	{"-iregex", false, parse_regex, REG_ICASE},
-	{"-iwholename", false, parse_path, true},
-	{"-links", false, parse_links},
-	{"-lname", false, parse_lname, false},
-	{"-ls", false, parse_ls},
-	{"-maxdepth", false, parse_depth_limit, false},
-	{"-mindepth", false, parse_depth_limit, true},
-	{"-mmin", false, parse_time, BFS_STAT_MTIME, MINUTES},
-	{"-mnewer", false, parse_newer, BFS_STAT_MTIME},
-	{"-mount", false, parse_mount},
-	{"-mtime", false, parse_time, BFS_STAT_MTIME, DAYS},
-	{"-name", false, parse_name, false},
-	{"-newer", false, parse_newer, BFS_STAT_MTIME},
-	{"-newer", true, parse_newerxy},
-	{"-nocolor", false, parse_color, false},
-	{"-nogroup", false, parse_nogroup},
-	{"-nohidden", false, parse_nohidden},
-	{"-noignore_readdir_race", false, parse_ignore_races, false},
-	{"-noleaf", false, parse_noleaf},
-	{"-not"},
-	{"-nouser", false, parse_nouser},
-	{"-nowarn", false, parse_warn, false},
-	{"-o"},
-	{"-ok", false, parse_exec, BFS_EXEC_CONFIRM},
-	{"-okdir", false, parse_exec, BFS_EXEC_CONFIRM | BFS_EXEC_CHDIR},
-	{"-or"},
-	{"-path", false, parse_path, false},
-	{"-perm", false, parse_perm},
-	{"-print", false, parse_print},
-	{"-print0", false, parse_print0},
-	{"-printf", false, parse_printf},
-	{"-printx", false, parse_printx},
-	{"-prune", false, parse_prune},
-	{"-quit", false, parse_quit},
-	{"-readable", false, parse_access, R_OK},
-	{"-regex", false, parse_regex, 0},
-	{"-regextype", false, parse_regextype},
-	{"-rm", false, parse_delete},
-	{"-samefile", false, parse_samefile},
-	{"-size", false, parse_size},
-	{"-sparse", false, parse_sparse},
-	{"-true", false, parse_const, true},
-	{"-type", false, parse_type, false},
-	{"-uid", false, parse_user},
-	{"-used", false, parse_used},
-	{"-user", false, parse_user},
-	{"-version", false, parse_version},
-	{"-warn", false, parse_warn, true},
-	{"-wholename", false, parse_path, false},
-	{"-writable", false, parse_access, W_OK},
-	{"-x", false, parse_mount},
-	{"-xdev", false, parse_mount},
-	{"-xtype", false, parse_type, true},
 	{"--"},
-	{"--help", false, parse_help},
-	{"--version", false, parse_version},
+	{"--help", parse_help},
+	{"--version", parse_version},
+	{"-Bmin", parse_time, BFS_STAT_BTIME, MINUTES},
+	{"-Bnewer", parse_newer, BFS_STAT_BTIME},
+	{"-Btime", parse_time, BFS_STAT_BTIME, DAYS},
+	{"-D", parse_debug},
+	{"-E", parse_regex_extended},
+	{"-H", parse_follow, BFTW_COMFOLLOW, false},
+	{"-L", parse_follow, BFTW_LOGICAL, false},
+	{"-O", parse_optlevel, 0, 0, true},
+	{"-P", parse_follow, 0, false},
+	{"-S", parse_search_strategy},
+	{"-X", parse_xargs_safe},
+	{"-a"},
+	{"-acl", parse_acl},
+	{"-amin", parse_time, BFS_STAT_ATIME, MINUTES},
+	{"-and"},
+	{"-anewer", parse_newer, BFS_STAT_ATIME},
+	{"-atime", parse_time, BFS_STAT_ATIME, DAYS},
+	{"-capable", parse_capable},
+	{"-cmin", parse_time, BFS_STAT_CTIME, MINUTES},
+	{"-cnewer", parse_newer, BFS_STAT_CTIME},
+	{"-color", parse_color, true},
+	{"-ctime", parse_time, BFS_STAT_CTIME, DAYS},
+	{"-d", parse_depth},
+	{"-daystart", parse_daystart},
+	{"-delete", parse_delete},
+	{"-depth", parse_depth_n},
+	{"-empty", parse_empty},
+	{"-exec", parse_exec, 0},
+	{"-execdir", parse_exec, BFS_EXEC_CHDIR},
+	{"-executable", parse_access, X_OK},
+	{"-exit", parse_exit},
+	{"-f", parse_f},
+	{"-false", parse_const, false},
+	{"-fls", parse_fls},
+	{"-follow", parse_follow, BFTW_LOGICAL, true},
+	{"-fprint", parse_fprint},
+	{"-fprint0", parse_fprint0},
+	{"-fprintf", parse_fprintf},
+	{"-fstype", parse_fstype},
+	{"-gid", parse_group},
+	{"-group", parse_group},
+	{"-help", parse_help},
+	{"-hidden", parse_hidden},
+	{"-ignore_readdir_race", parse_ignore_races, true},
+	{"-ilname", parse_lname, true},
+	{"-iname", parse_name, true},
+	{"-inum", parse_inum},
+	{"-ipath", parse_path, true},
+	{"-iregex", parse_regex, REG_ICASE},
+	{"-iwholename", parse_path, true},
+	{"-links", parse_links},
+	{"-lname", parse_lname, false},
+	{"-ls", parse_ls},
+	{"-maxdepth", parse_depth_limit, false},
+	{"-mindepth", parse_depth_limit, true},
+	{"-mmin", parse_time, BFS_STAT_MTIME, MINUTES},
+	{"-mnewer", parse_newer, BFS_STAT_MTIME},
+	{"-mount", parse_mount},
+	{"-mtime", parse_time, BFS_STAT_MTIME, DAYS},
+	{"-name", parse_name, false},
+	{"-newer", parse_newer, BFS_STAT_MTIME},
+	{"-newer", parse_newerxy, 0, 0, true},
+	{"-nocolor", parse_color, false},
+	{"-nogroup", parse_nogroup},
+	{"-nohidden", parse_nohidden},
+	{"-noignore_readdir_race", parse_ignore_races, false},
+	{"-noleaf", parse_noleaf},
+	{"-not"},
+	{"-nouser", parse_nouser},
+	{"-nowarn", parse_warn, false},
+	{"-o"},
+	{"-ok", parse_exec, BFS_EXEC_CONFIRM},
+	{"-okdir", parse_exec, BFS_EXEC_CONFIRM | BFS_EXEC_CHDIR},
+	{"-or"},
+	{"-path", parse_path, false},
+	{"-perm", parse_perm},
+	{"-print", parse_print},
+	{"-print0", parse_print0},
+	{"-printf", parse_printf},
+	{"-printx", parse_printx},
+	{"-prune", parse_prune},
+	{"-quit", parse_quit},
+	{"-readable", parse_access, R_OK},
+	{"-regex", parse_regex, 0},
+	{"-regextype", parse_regextype},
+	{"-rm", parse_delete},
+	{"-samefile", parse_samefile},
+	{"-size", parse_size},
+	{"-sparse", parse_sparse},
+	{"-true", parse_const, true},
+	{"-type", parse_type, false},
+	{"-uid", parse_user},
+	{"-unique", parse_unique},
+	{"-used", parse_used},
+	{"-user", parse_user},
+	{"-version", parse_version},
+	{"-warn", parse_warn, true},
+	{"-wholename", parse_path, false},
+	{"-writable", parse_access, W_OK},
+	{"-x", parse_mount},
+	{"-xattr", parse_xattr},
+	{"-xdev", parse_mount},
+	{"-xtype", parse_type, true},
 	{0},
 };
 
@@ -3003,6 +3211,21 @@ void dump_cmdline(const struct cmdline *cmdline, bool verbose) {
 
 	cfprintf(cerr, "${ex}%s${rs} ", cmdline->argv[0]);
 
+	const char *strategy = NULL;
+	switch (cmdline->strategy) {
+	case BFTW_BFS:
+		strategy = "bfs";
+		break;
+	case BFTW_DFS:
+		strategy = "dfs";
+		break;
+	case BFTW_IDS:
+		strategy = "ids";
+		break;
+	}
+	assert(strategy);
+	cfprintf(cerr, "${cyn}-S${rs} ${bld}%s${rs} ", strategy);
+
 	if (cmdline->flags & BFTW_LOGICAL) {
 		cfprintf(cerr, "${cyn}-L${rs} ");
 	} else if (cmdline->flags & BFTW_COMFOLLOW) {
@@ -3032,12 +3255,13 @@ void dump_cmdline(const struct cmdline *cmdline, bool verbose) {
 		cfprintf(cerr, " ");
 	}
 
-	for (struct root *root = cmdline->roots; root; root = root->next) {
-		char c = root->path[0];
+	for (size_t i = 0; i < cmdline->npaths; ++i) {
+		const char *path = cmdline->paths[i];
+		char c = path[0];
 		if (c == '-' || c == '(' || c == ')' || c == '!' || c == ',') {
 			cfprintf(cerr, "${cyn}-f${rs} ");
 		}
-		cfprintf(cerr, "${mag}%s${rs} ", root->path);
+		cfprintf(cerr, "${mag}%s${rs} ", path);
 	}
 
 	if (cmdline->cout->colors) {
@@ -3060,6 +3284,9 @@ void dump_cmdline(const struct cmdline *cmdline, bool verbose) {
 	if (cmdline->maxdepth != INT_MAX) {
 		cfprintf(cerr, "${blu}-maxdepth${rs} ${bld}%d${rs} ", cmdline->maxdepth);
 	}
+	if (cmdline->unique) {
+		cfprintf(cerr, "${blu}-unique${rs} ");
+	}
 
 	dump_expr(cerr, cmdline->expr, verbose);
 
@@ -3110,22 +3337,27 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 	}
 
 	cmdline->argv = NULL;
-	cmdline->roots = NULL;
+	cmdline->paths = NULL;
+	cmdline->npaths = 0;
 	cmdline->colors = NULL;
 	cmdline->cout = NULL;
 	cmdline->cerr = NULL;
 	cmdline->mtab = NULL;
+	cmdline->mtab_error = 0;
 	cmdline->mindepth = 0;
 	cmdline->maxdepth = INT_MAX;
 	cmdline->flags = BFTW_RECOVER;
+	cmdline->strategy = BFTW_BFS;
 	cmdline->optlevel = 3;
 	cmdline->debug = 0;
 	cmdline->xargs_safe = false;
 	cmdline->ignore_races = false;
+	cmdline->unique = false;
 	cmdline->expr = &expr_true;
-	cmdline->open_files = NULL;
 	cmdline->nopen_files = 0;
 
+	trie_init(&cmdline->open_files);
+
 	static char* default_argv[] = {"bfs", NULL};
 	if (argc < 1) {
 		argc = 1;
@@ -3141,25 +3373,37 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 		cmdline->argv[i] = argv[i];
 	}
 
+	enum use_color use_color = COLOR_AUTO;
+	if (getenv("NO_COLOR")) {
+		// https://no-color.org/
+		use_color = COLOR_NEVER;
+	}
+
 	cmdline->colors = parse_colors(getenv("LS_COLORS"));
-	cmdline->cout = cfdup(stdout, cmdline->colors);
-	cmdline->cerr = cfdup(stderr, cmdline->colors);
+	cmdline->cout = cfdup(stdout, use_color ? cmdline->colors : NULL);
+	cmdline->cerr = cfdup(stderr, use_color ? cmdline->colors : NULL);
 	if (!cmdline->cout || !cmdline->cerr) {
 		perror("cfdup()");
 		goto fail;
 	}
 
-	bool stderr_tty = cmdline->cerr->colors;
+	cmdline->mtab = parse_bfs_mtab();
+	if (!cmdline->mtab) {
+		cmdline->mtab_error = errno;
+	}
+
 	bool stdin_tty = isatty(STDIN_FILENO);
+	bool stdout_tty = isatty(STDOUT_FILENO);
+	bool stderr_tty = isatty(STDERR_FILENO);
 
 	struct parser_state state = {
 		.cmdline = cmdline,
 		.argv = cmdline->argv + 1,
 		.command = cmdline->argv[0],
-		.roots_tail = &cmdline->roots,
 		.regex_flags = 0,
-		.interactive = stderr_tty && stdin_tty,
-		.use_color = COLOR_AUTO,
+		.stdout_tty = stdout_tty,
+		.interactive = stdin_tty && stderr_tty,
+		.use_color = use_color,
 		.implicit_print = true,
 		.warn = stdin_tty,
 		.non_option_seen = false,
@@ -3169,6 +3413,11 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 		.prune_arg = NULL,
 	};
 
+	if (strcmp(xbasename(state.command), "find") == 0) {
+		// Operate depth-first when invoked as "find"
+		cmdline->strategy = BFTW_DFS;
+	}
+
 	if (parse_gettime(&state.now) != 0) {
 		goto fail;
 	}
@@ -3186,12 +3435,17 @@ struct cmdline *parse_cmdline(int argc, char *argv[]) {
 		goto fail;
 	}
 
-	if (!cmdline->roots) {
+	if (cmdline->npaths == 0) {
 		if (parse_root(&state, ".") != 0) {
 			goto fail;
 		}
 	}
 
+	if ((cmdline->flags & BFTW_LOGICAL) && !cmdline->unique) {
+		// We need bftw() to detect cycles unless -unique does it for us
+		cmdline->flags |= BFTW_DETECT_CYCLES;
+	}
+
 	if (cmdline->debug & DEBUG_TREE) {
 		dump_cmdline(cmdline, false);
 	}
