@@ -1,6 +1,6 @@
 /****************************************************************************
  * bfs                                                                      *
- * Copyright (C) 2018 Tavian Barnes <tavianator@tavianator.com>             *
+ * Copyright (C) 2018-2019 Tavian Barnes <tavianator@tavianator.com>        *
  *                                                                          *
  * Permission to use, copy, modify, and/or distribute this software for any *
  * purpose with or without fee is hereby granted.                           *
@@ -82,10 +82,9 @@ const char *bfs_stat_field_name(enum bfs_stat_field field) {
 /**
  * Check if we should retry a failed stat() due to a potentially broken link.
  */
-static bool bfs_stat_retry(int ret, int at_flags, enum bfs_stat_flag flags) {
+static bool bfs_stat_retry(int ret, enum bfs_stat_flag flags) {
 	return ret != 0
-		&& !(at_flags & AT_SYMLINK_NOFOLLOW)
-		&& (flags & BFS_STAT_BROKEN_OK)
+		&& (flags & (BFS_STAT_NOFOLLOW | BFS_STAT_TRYFOLLOW)) == BFS_STAT_TRYFOLLOW
 		&& is_nonexistence_error(errno);
 }
 
@@ -144,7 +143,7 @@ static int bfs_stat_impl(int at_fd, const char *at_path, int at_flags, enum bfs_
 	struct stat statbuf;
 	int ret = fstatat(at_fd, at_path, &statbuf, at_flags);
 
-	if (bfs_stat_retry(ret, at_flags, flags)) {
+	if (bfs_stat_retry(ret, flags)) {
 		at_flags |= AT_SYMLINK_NOFOLLOW;
 		ret = fstatat(at_fd, at_path, &statbuf, at_flags);
 	}
@@ -183,7 +182,7 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, enum bfs
 	struct statx xbuf;
 	int ret = bfs_statx(at_fd, at_path, at_flags, mask, &xbuf);
 
-	if (bfs_stat_retry(ret, at_flags, flags)) {
+	if (bfs_stat_retry(ret, flags)) {
 		at_flags |= AT_SYMLINK_NOFOLLOW;
 		ret = bfs_statx(at_fd, at_path, at_flags, mask, &xbuf);
 	}
@@ -274,7 +273,10 @@ static int bfs_statx_impl(int at_fd, const char *at_path, int at_flags, enum bfs
 
 #endif // HAVE_BFS_STATX
 
-int bfs_stat(int at_fd, const char *at_path, int at_flags, enum bfs_stat_flag flags, struct bfs_stat *buf) {
+/**
+ * Allows calling stat with custom at_flags.
+ */
+static int bfs_stat_explicit(int at_fd, const char *at_path, int at_flags, enum bfs_stat_flag flags, struct bfs_stat *buf) {
 #if HAVE_BFS_STATX
 	static bool has_statx = true;
 
@@ -291,12 +293,21 @@ int bfs_stat(int at_fd, const char *at_path, int at_flags, enum bfs_stat_flag fl
 	return bfs_stat_impl(at_fd, at_path, at_flags, flags, buf);
 }
 
-int bfs_fstat(int fd, struct bfs_stat *buf) {
+int bfs_stat(int at_fd, const char *at_path, enum bfs_stat_flag flags, struct bfs_stat *buf) {
+	int at_flags = 0;
+	if (flags & BFS_STAT_NOFOLLOW) {
+		at_flags |= AT_SYMLINK_NOFOLLOW;
+	}
+
+	if (at_path) {
+		return bfs_stat_explicit(at_fd, at_path, at_flags, flags, buf);
+	}
+
 #ifdef AT_EMPTY_PATH
 	static bool has_at_ep = true;
-
 	if (has_at_ep) {
-		int ret = bfs_stat(fd, "", AT_EMPTY_PATH, 0, buf);
+		at_flags |= AT_EMPTY_PATH;
+		int ret = bfs_stat_explicit(at_fd, "", at_flags, flags, buf);
 		if (ret != 0 && errno == EINVAL) {
 			has_at_ep = false;
 		} else {
@@ -306,11 +317,12 @@ int bfs_fstat(int fd, struct bfs_stat *buf) {
 #endif
 
 	struct stat statbuf;
-	int ret = fstat(fd, &statbuf);
-	if (ret == 0) {
+	if (fstat(at_fd, &statbuf) == 0) {
 		bfs_stat_convert(&statbuf, buf);
+		return 0;
+	} else {
+		return -1;
 	}
-	return ret;
 }
 
 const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_field field) {
@@ -334,3 +346,8 @@ const struct timespec *bfs_stat_time(const struct bfs_stat *buf, enum bfs_stat_f
 		return NULL;
 	}
 }
+
+void bfs_stat_id(const struct bfs_stat *buf, bfs_file_id *id) {
+	memcpy(*id, &buf->dev, sizeof(buf->dev));
+	memcpy(*id + sizeof(buf->dev), &buf->ino, sizeof(buf->ino));
+}
